;;;; ITUNES SONG LIST CONVERTER MODULE
;;;; This program takes the xml file exported by iTunes when using the
;;;; Export Song List option from the File menu and converts it to an
;;;; simple xhtml table.

;;; Copyright (c) 2006, Aaron Hsu
;;; All rights reserved.
;;;
;;; Redistribution and use in source and binary forms, with or without
;;; modification, are permitted provided that the following conditions are
;;; met:
;;;
;;;     Redistributions of source code must retain the above copyright
;;;     notice, this list of conditions and the following disclaimer.
;;;
;;;     Redistributions in binary form must reproduce the above copyright
;;;     notice, this list of conditions and the following disclaimer in
;;;     the documentation and/or other materials provided with the
;;;     distribution.
;;;
;;;     Neither the name of the original developer nor the names of its
;;;     contributors may be used to endorse or promote products derived
;;;     from this software without specific prior written permission.
;;;
;;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(module
 itunes-songlist-converter (itsl->html)

 ;; get-entry-line : iport -> string
 ;; Returns the next entry-line from the iport, removing tabs.
 (define (get-entry-line iport)
   (let loop ((c (read-char iport))
	      (s (string)))
     (case c
       ((#\tab)
	(loop (read-char iport)
	      s))
       ((#\return)
	(let ((c (read-char iport)))
	  (if (char=? c #\newline)
	      s
	      (loop (read-char iport)
		    (string-append s (string #\return c))))))
       (else (loop (read-char iport)
		   (string-append s (string c)))))))

 ;; find-tag : string close? end? -> integer
 ;; Rreads characters from string, finding an xml type tag. If close is
 ;; set to #f then it will find an opening tag, and if type is set
 ;; to #t, it will find a closing tag. End? is a boolean, which, if
 ;; set, with cause find-tag to return the end of the tag as the index,
 ;; rather than the beginning of the tag.
 (define (find-tag s t e)
   (if t ; Searching for close tag?
       (if e ; Want end of index
	   (let loop ((i 0)
		      (f #f))
	     (cond ((> (+ i 2)
		       (string-length s))
		    (string-length s))
		   ((string=? (substring s i (+ i 2))
			      "</")
		    (loop (+ i 1)
			  #t))
		   ((and f
			 (char=? (string-ref s i)
				 #\>))
		    (+ i 1))
		   (else (loop (+ i 1)
			       f))))
	   (let loop ((i 0))
	     (cond ((> (+ i 2)
		       (string-length s))
		    (- (string-length s)
		       2))
		   ((string=? (substring s i (+ i 2))
			      "</")
		    i)
		   (else (loop (+ i 1))))))
       (if e ; Want end of index
	   (let loop ((i 0)
		      (f #f)
		      (c 0))
	     (cond ((char=? (string-ref s i)
			    #\<)
		    (loop (+ i 1)
			  #t
			  (+ i 1)))
		   ((and f
			 (char=? (string-ref s i)
				 #\>))
		    (+ i 1))
		   ((and f
			 (char=? (string-ref s i)
				 #\/))
		    c)
		   (else (loop (+ i 1)
			       f
			       c))))
	   (let loop ((i 0))
	     (case (string-ref s i)
	       ((#\<)
		i)
	       (else (loop (+ i 1))))))))

 ;; Takes a string containing key information and data like
 ;; "<key>...</key><datatag>...</datatag>" and converts it to a pair
 ;; whose car is a symbol equal to the contents of the key tag, and
 ;; whose cdr is a string representing the contents in datatag.
 (define (get-key-pair key-line)
   (cons (string->symbol (substring key-line
				    (find-tag key-line #f #t)
				    (find-tag key-line #t #f)))
	 (let ((s (substring key-line
			     (find-tag key-line #t #t)
			     (string-length key-line))))
	   (substring s
		      (find-tag s #f #t)
		      (find-tag s #t #f)))))
 
 ;; get-key-line : iport -> string
 ;; Takes an input port, reads it for key-lines. If it finds a
 ;; keyline, it returns the complete keyline (spanning newlines), but
 ;; if it does not find a key, then it returns the next line in the
 ;; port.
 (define (get-key-line iport)
   (define (get-tag iport)
     (let loop ((c (read-char iport))
		(s (string)))
       (case c
	 ((#\/)
	  (let ((c2 (read-char iport)))
	    (if (char=? c2 #\>)
		(string-append s "/>")
		(loop (read-char iport)
		      (string-append s (string #\/ c2))))))
	 ((#\>)
	  (string-append s ">"))
	 (else (loop (read-char iport)
		     (string-append s (string c)))))))
   (define (get-key-end iport)
     (let loop ((c (read-char iport))
		(s (string)))
       (if (char=? c #\<)
	   (let ((s2 (string-append "<" (get-tag iport))))
	     (if (string=? s2 "</key>")
		 (string-append s s2)
		 (loop (read-char iport)
		       (string-append s s2))))
	   (loop (read-char iport)
		 (string-append s (string c))))))
   (define (get-complete-tag iport)
     (let loop ((c (read-char iport))
		(s (string)))
       (if (char=? c #\<)
	   (let ((tag (string-append "<" (get-tag iport))))
	     (if (char=? (string-ref tag (- (string-length tag)
					    2))
			 #\/)
		 (begin (read-char iport)
			(read-char iport)
			(string-append s tag))
		 (let ((etag (string-append "</" (substring
						  tag
						  1
						  (string-length
						   tag)))))
		   (let loop2 ((c (read-char iport))
			       (s2 (string)))
		     (if (char=? c #\<)
			 (let ((tag2 (string-append "<" (get-tag iport))))
			   (if (string=? tag2 etag)
			       (begin (read-char iport)
				      (read-char iport)
				      (string-append s tag s2 tag2))
			       (loop2 (read-char iport)
				      (string-append s2 tag2))))
			 (loop2 (read-char iport)
				(string-append s2 (string c))))))))
	   (loop (read-char iport)
		 (string-append s (string c))))))
   (let loop ((c (read-char iport))
	      (s (string)))
     (case c
       ((#\tab)
	(loop (read-char iport)
	      s))
       ((#\return)
	(let ((c (read-char iport)))
	  (if (char=? c #\newline)
	      s
	      (loop (read-char iport)
		    (string-append s (string #\return c))))))
       ((#\<)
	(let ((tag (string-append "<" (get-tag iport))))
	  (if (string=? "<key>" tag)
	      (let ((keytag (string-append tag (get-key-end iport))))
		(string-append s keytag (get-complete-tag iport)))
	      (loop (read-char iport)
		    (string-append s tag)))))
       (else (loop (read-char iport)
		   (string-append s (string c)))))))
  
 ;; get-entry : iport -> entry-list
 ;; reads from iport a <dict>...</dict> form and generates a list of
 ;; pairs, each pair containing a symbol representing the key, and a
 ;; string representing the contents after the key.
 (define (get-entry iport)
   (let loop ((key-line (get-key-line iport))
	      (entry-list (list)))
     (cond ((string=? "<key>" (substring key-line 0 5))
	    (loop (get-key-line iport)
		  (append entry-list (list (get-key-pair key-line)))))
	   ((string=? "</dict>" key-line)
	    entry-list)
	   (else (loop (get-key-line iport)
		       entry-list)))))

 ;; advance-to-entries : iport -> undefined
 ;; Reads the input port forward until it finds a Tracks key entry.
 (define (advance-to-entries iport)
   (let loop ((entry-line (get-entry-line iport)))
     (if (not (string=? "<key>Tracks</key>" entry-line))
	 (loop (get-entry-line iport)))))

 ;; get-tracks : input-port -> track-list
 ;; Reads the xml song list and converts all tracks to list form, and
 ;; returns a list of entry-lists. An entry list is a list of pairs,
 ;; each pair consisting of a symbol representing the field name, and a
 ;; string representing the contents of that field.
 (define (get-tracks iport)
   (advance-to-entries iport)
   (let loop ((track-list (list)))
     (define entry-line (get-entry-line iport))
     (cond ((string=? "<key>" (substring entry-line 0 5))
	    (loop (append track-list (list (get-entry iport)))))
	   ((string=? "</dict>" entry-line)
	    (close-input-port iport)
	    track-list)
	   (else (loop track-list)))))

 ;; itsl->html : input-file output-file field-list -> boolean
 ;; Takes the input file, parses it and extracts all the data according
 ;; to the desired fields and then writes it to the output file.
 ;;
 ;; field-list is a list of symbols representing the fields you want in
 ;; the columns.
 (define (itsl->html ifile ofile fl)
   (let ((oport (open-output-file ofile '|replace|))
	 (tl (call-with-input-file ifile get-tracks)))
     (display "<table>" oport)
     (newline oport)
     (display "  <tr>" oport)
     (newline oport)
     (let loop ((fl fl))
       (if (not (null? fl))
	   (begin (display "    <th>" oport)
		  (newline oport)
		  (display "      " oport)
		  (display (symbol->string (car fl))
			   oport)
		  (newline oport)
		  (display "    </th>" oport)
		  (newline oport)
		  (loop (cdr fl)))))
     (display "  </tr>" oport)
     (newline oport)
     (let entryloop ((tl tl))
       (if (not (null? tl))
	   (begin (display "  <tr>" oport)
		  (newline oport)
		  (let fieldloop ((fl fl))
		    (if (not (null? fl))
			(begin (display "    <td>" oport)
			       (newline oport)
			       (display "      " oport)
			       (let ((item (assq (car fl)
						 (car tl))))
				 (if (pair? item)
				     (display (cdr item)
					      oport)))
			       (newline oport)
			       (display "    </td>" oport)
			       (newline oport)
			       (fieldloop (cdr fl)))))
		  (display "  </tr>" oport)
		  (newline oport)
		  (entryloop (cdr tl)))))
     (display "</table>" oport)
     (newline oport)
     (close-output-port oport))))